{$A+,B-,E+,F-,G+,I-,N+,P-,Q-,R-,S-,T-,V-,X+}

program FruitPopper;

uses GDGfx, GDKeybrd, GDIFF, FixedP, Toolbox, GDPCX;

const
  BACKBUFFER_LAYER           = 1;
  SCREEN_MAP_LAYER           = 2;

  SCREEN_MAP_WIDTH           = 20;
  SCREEN_MAP_HEIGHT          = 11;

  PLAYER_1_TILE_START        = 40;
  PLAYER_2_TILE_START        = 60;

  { every tile beginning with this index should be
    considered 'solid' for collision-purposes }
  SOLID_TILE_START           = 70;

  PLAYER_FRICTION            = FP_0_2;
  PLAYER_WALK_SPEED          = FP_0_6;

type
  Direction = (South, North, East, West);
  Animation = (Idle, Walking, Stabbing, Victory, Defeat);

  AnimationDesc = record
    frames     : array[0..5] of word; { 6 frames total ever ... }
    count      : word; { number of frames }
    delay      : word; { frame timer/delay between switching to next frame }
    loops      : bytebool; { true = animation loops }
    base       : word; { spritesheet index of first frame. if
                         multi-directional, this should be the index of the
                         first frame for the south direction }
    dirLength  : word; { number of frames from the start of one direction
                         to the start of the next direction. if not a
                         multi-directional animation, this must be set to 0 }
  end;
  PAnimationDesc = ^AnimationDesc;

  AnimationState = record
    complete   : bytebool; { true = current animation sequence is complete }
    frameIndex : word; { index of the current animation sequence frame }
    time       : word; { current frame timer }
  end;
  PAnimationState = ^AnimationState;

  ScreenMap = array[0..((SCREEN_MAP_WIDTH*SCREEN_MAP_HEIGHT)-1)] of byte;


  PlayerState = record
    x, y           : fixed;
    velX, velY     : fixed;
    forceX, forceY : fixed;
    dir            : Direction;
    state          : Animation;
    animation      : AnimationState;
    spriteOffset   : word; { spritesheet offset to start of char sprites }
  end;

const

  { TODO: load this from file(s) ... }
  map : ScreenMap = (
    74,72,72,72,72,72,72,72,72,72,72,72,72,72,72,72,72,72,72,75,
    73,26,24,38,24,38,24,39,16,16,16,16,40,24,38,24,38,24,27,73,
    73,25,70,25,70,25,70,18,15,15,15,15,19,70,25,70,25,70,25,73,
    73,44,24,54,24,33,16,33,15,15,15,15,32,16,32,24,54,24,47,73,
    73,25,70,25,70,18,15,15,15,15,15,15,15,15,19,70,25,70,25,73,
    73,45,24,54,24,36,15,15,15,15,15,15,15,15,37,24,54,24,48,73,
    73,25,70,25,70,18,15,15,15,15,15,15,15,15,19,70,25,70,25,73,
    73,44,24,54,24,31,17,31,15,15,15,15,30,17,30,24,54,24,47,73,
    73,25,70,25,70,25,70,18,15,15,15,15,19,70,25,70,25,70,25,73,
    73,22,24,41,24,41,24,42,17,17,17,17,43,24,41,24,41,24,23,73,
    76,72,72,72,72,72,72,72,72,72,72,72,72,72,72,72,72,72,72,77
  );

  isMapDirty : boolean = false;

var
  pal              : Palette;
  tiles            : array[0..119] of PBitmap;
  sprites          : array[0..119] of PBitmap;

  player           : PlayerState;
  animations       : array[0..ord(Defeat)] of AnimationDesc;


function LoadTilesAndSprites(filename : string) : boolean;
var
  i, x, y, offset : integer;
label return;
begin
  LoadTilesAndSprites := false;

  UseLayer(BACKBUFFER_LAYER);
  if LoadIFF(filename, @pal) <> IFFOk then goto return;

  { environment tiles are on the left }
  offset := 0;
  i      := 0;
  for y := 0 to 11 do begin
    for x := 0 to 9 do begin
      AllocBitmap(16, 16, tiles[i]);
      GrabBitmap((x*16)+offset, y*16, 16, 16, tiles[i]);
      inc(i);
    end;
  end;

  { sprites are on the right }
  offset := 160;
  i      := 0;
  for y := 0 to 11 do begin
    for x := 0 to 9 do begin
      AllocBitmap(16, 16, sprites[i]);
      GrabBitmap((x*16)+offset, y*16, 16, 16, sprites[i]);
      inc(i);
    end;
  end;

  LoadTilesAndSprites := true;

return:
  UseLayer(SCREEN_LAYER);
end;

function LoadScreenMap(filename : string) : boolean;
begin
  { TODO }
  isMapDirty := true;
  LoadScreenMap := true;
end;

procedure DrawScreenMap;
var
  x, y           : word;
  currentMapTile : ^byte;
  oldLayer       : integer;
begin
  if isMapDirty then begin
    { re-render map from screen map tile data if something has changed
      that map data }
    oldLayer := GetBoundLayerIndex;
    UseLayer(SCREEN_MAP_LAYER);

    currentMapTile := @map[0];
    for y := 0 to 10 do begin
      for x := 0 to 19 do begin
        Blitf(x*16, y*16, tiles[currentMapTile^]);
        inc(currentMapTile);
      end;
    end;

    UseLayer(oldLayer);
    isMapDirty := false;
  end else begin
    { otherwise, just copy the pre-rendered map }
    CopyLayer(SCREEN_MAP_LAYER);
  end;
end;

procedure SetPlayerAnimation(var player : PlayerState; animation : Animation);
begin
  if animation = player.state then exit;
  player.state := animation;
  player.animation.complete := false;
  player.animation.frameIndex := 0;
  player.animation.time := 0;
end;

function IsPlayerStopped(const player : PlayerState) : boolean;
{ returns true if the player's movement velocity is slow enough that they
  could be considered stopped. this does not take into account the player's
  force velocity }
const SLOW_ENOUGH = FP_0_1;
begin
  with player do begin
    IsPlayerStopped := (abs(velX) < SLOW_ENOUGH)
                        and (abs(velY) < SLOW_ENOUGH);
  end;
end;

function IsMapCollision(x, y : integer) : boolean;
const
  ENTITY_SIZE = 16;
  TILE_SIZE = 16;
  MAP_RIGHT = SCREEN_MAP_WIDTH - 1;
  MAP_BOTTOM = SCREEN_MAP_HEIGHT - 1;
var
  left, right, top, bottom : integer;
  check_x, check_y         : integer;
begin
  IsMapCollision := false;

  left := x div TILE_SIZE;
  right := (x + ENTITY_SIZE-1) div TILE_SIZE;
  top := y div TILE_SIZE;
  bottom := (y + ENTITY_SIZE-1) div TILE_SIZE;

  if left < 0 then left := 0;
  if right > MAP_RIGHT then right := MAP_RIGHT;
  if top < 0 then top := 0;
  if bottom > MAP_BOTTOM then bottom := MAP_BOTTOM;

  for check_y := top to bottom do begin
    for check_x := left to right do begin
      if map[(check_y * SCREEN_MAP_WIDTH) + check_x] >= SOLID_TILE_START then begin
        IsMapCollision := true;
        exit;
      end;
    end;
  end;
end;

function MovePlayer(var player : PlayerState) : boolean;
{ updated the player's X and Y position based on their current movement
  velocity and any force velocity. checks for map collisions and prevents
  movement along the X and/or Y axis if any collisions are found. returns
  true if the player collided against something. }
var
  speedX, speedY : fixed;
begin
  MovePlayer := false;

  with player do begin
    speedX := velX + forceX;
    speedY := velY + forceY;

    inc(x, speedX);
    if IsMapCollision(FixToInt(x), FixToInt(y)) then begin
      MovePlayer := true;
      dec(x, speedX);
    end;

    inc(y, speedY);
    if IsMapCollision(FixToInt(x), FixToInt(y)) then begin
      MovePlayer := true;
      dec(y, speedY);
    end;
  end;
end;

procedure UpdatePlayer(var player : PlayerState);
var
  animation : ^AnimationDesc;
begin
  with player do begin
    { move player in the direction of their velocity and any combined force
      that is currently being applied to them. also handles collision. }
    MovePlayer(player);

    { slow velocity down by friction }
    velX := FixMul(velX, PLAYER_FRICTION);
    velY := FixMul(velY, PLAYER_FRICTION);

    { degrade strength of any combined force }
    { TODO: use something other than friction for this??? }
    forceX := FixMul(forceX, PLAYER_FRICTION);
    forceY := FixMul(forceY, PLAYER_FRICTION);

    { set player idle/walking based on their velocity.
      note that this check ignores their force velocity! }
    if IsPlayerStopped(player) then
      SetPlayerAnimation(player, Idle)
    else
      SetPlayerAnimation(player, Walking);
  end;

  with player.animation do begin
    if not complete then begin
      animation := @animations[ord(player.state)];

      inc(time);
      if time >= animation^.delay then begin
        { move to next frame in the current animation sequence }
        time := 0;
        if frameIndex < (animation^.count-1) then begin
          inc(frameIndex);
        end else begin
          { we're at the last frame in the current animation sequence }
          if not animation^.loops then begin
            complete := true;
          end else
            frameIndex := 0;
        end;
      end;
    end;
  end;
end;

function GetAnimationFrame(const state : PAnimationState;
                           const desc : PAnimationDesc;
                           dir : Direction) : word;
{ returns the current spritesheet bitmap index that should be blitted
  to draw an entity based on it's animation state and facing direction }
begin
  with desc^ do begin
    GetAnimationFrame := frames[state^.frameIndex]
                          + (ord(dir) * dirLength)
                          + base;
  end;
end;

procedure DrawPlayer(var player : PlayerState);
var
  index : word;
begin
  index := GetAnimationFrame(
    @player.animation,
    @animations[ord(player.state)],
    player.dir
  );

  BlitSprite(
    FixToInt(player.x),
    FixToInt(player.y),
    sprites[index + player.spriteOffset]
  );
end;

procedure MainLoop;
var
  frames, fps      : integer;
  ticks, lastTicks : fixed;
begin
  frames := 0;
  fps := 0;
  ticks := BiosTimeSecondsFP;
  lastTicks := ticks;

  isMapDirty := true;

  while not (Keys[KEY_ESC]) do begin

    if Keys[KEY_LEFT] then begin
      player.velX := player.velX - PLAYER_WALK_SPEED;
      player.dir := West;
    end;
    if Keys[KEY_RIGHT] then begin
      player.velX := player.velX + PLAYER_WALK_SPEED;
      player.dir := East;
    end;
    if Keys[KEY_UP] then begin
      player.velY := player.velY - PLAYER_WALK_SPEED;
      player.dir := North;
    end;
    if Keys[KEY_DOWN] then begin
      player.velY := player.velY + PLAYER_WALK_SPEED;
      player.dir := South;
    end;

    { update state }
    UpdatePlayer(player);

    { render }
    DrawScreenMap;
    DrawPlayer(player);

    { update fps stats }
    inc(frames);
    ticks := BiosTimeSecondsFP;
    if (ticks - lastTicks) >= FP_1 then begin
      fps := frames;
      frames := 0;
      lastTicks := ticks;
    end;

    PrintAt(0, 0); PrintInt(fps, 15);
    {
    PrintString(' ', 15);
    PrintSingleFmt(FixToFloat(player.velX), 2, 15); PrintString(' ', 15);
    PrintSingleFmt(FixToFloat(player.velY), 2, 15); PrintString(' ', 15);
    PrintSingleFmt(FixToFloat(player.forceX), 2, 15); PrintString(' ', 15);
    PrintSingleFmt(FixToFloat(player.forceY), 2, 15); PrintString(' ', 15);
    }

    PrintAt(0, 176); PrintString('Game status display goes here ...', 15);

    WaitForVsync;
    Flip(BACKBUFFER_LAYER);
  end;
end;

procedure InitStaticData;
begin

  animations[ord(Idle)].frames[0] := 0;
  animations[ord(Idle)].count := 1;
  animations[ord(Idle)].delay := 0;
  animations[ord(Idle)].loops := true;
  animations[ord(Idle)].base := 0;
  animations[ord(Idle)].dirLength := 3;

  animations[ord(Walking)].frames[0] := 1;
  animations[ord(Walking)].frames[1] := 0;
  animations[ord(Walking)].frames[2] := 2;
  animations[ord(Walking)].frames[3] := 0;
  animations[ord(Walking)].count := 4;
  animations[ord(Walking)].delay := 10;
  animations[ord(Walking)].loops := true;
  animations[ord(Walking)].base := 0;
  animations[ord(Walking)].dirLength := 3;

  animations[ord(Stabbing)].frames[0] := 0;
  animations[ord(Stabbing)].frames[1] := 0;
  animations[ord(Stabbing)].count := 2;
  animations[ord(Stabbing)].delay := 8;
  animations[ord(Stabbing)].base := 12;
  animations[ord(Stabbing)].dirLength := 1;
  animations[ord(Stabbing)].loops := false;


  animations[ord(Victory)].frames[0] := 0;
  animations[ord(Victory)].count := 1;
  animations[ord(Victory)].delay := 0;
  animations[ord(Victory)].loops := true;
  animations[ord(Victory)].base := 16;
  animations[ord(Victory)].dirLength := 0;

  animations[ord(Defeat)].frames[0] := 0;
  animations[ord(Defeat)].count := 1;
  animations[ord(Defeat)].delay := 0;
  animations[ord(Defeat)].loops := true;
  animations[ord(Defeat)].base := 17;
  animations[ord(Defeat)].dirLength := 0;
end;

label cleanUpAndExit;
begin
  InitStaticData;

  InitGraphics(2);
  InitKeyboard;

  Cls(0);
  PrintAt(0, 0);
  PrintString('Loading ...'#10, 15);

  if (not LoadTilesAndSprites('tiles.lbm')) then begin
    PrintString('Failed loading tiles and sprites', 15);
    WaitForAnyKey;
    goto cleanUpAndExit;
  end;

  PrintString('Ready', 15);

  LoadScreenMap('test.map');

  UseLayer(BACKBUFFER_LAYER);
  Cls(0);
  Flip(1);
  SetPalette(@pal);

  player.x := IntToFix(8*16);
  player.y := IntToFix(4*16);
  player.velX := 0;
  player.velY := 0;
  player.forceX := 0;
  player.forceY := 0;
  player.dir := South;
  player.spriteOffset := PLAYER_1_TILE_START;
  SetPlayerAnimation(player, Walking);

  MainLoop;

cleanUpAndExit:
  CloseKeyboard;
  CloseGraphics;

end.
